# flake8: noqa
from .af1 import Af1
from .ag3 import Ag3
from .amin1 import Amin1
from .anopheles import AnophelesDataResource, Region
from .pf7 import Pf7
from .pv4 import Pv4
from .util import SiteClass

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata  # type: ignore

# this will read version from pyproject.toml
__version__ = importlib_metadata.version(__name__)


def _ensure_xarray_zarr_backend():
    # workaround for https://github.com/malariagen/malariagen-data-python/issues/320
    # see also https://github.com/pydata/xarray/issues/7478

    from xarray.backends.plugins import list_engines
    from xarray.backends.zarr import ZarrBackendEntrypoint

    # zarr is a dependency, so will always be available if malariagen_data
    # is installed
    ZarrBackendEntrypoint.available = True

    # ensure xarray refreshes list of available engines
    list_engines.cache_clear()


_ensure_xarray_zarr_backend()



import json
from pathlib import Path
from typing import (
    IO,
    Any,
    Dict,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import bokeh.io
import numpy as np
import pandas as pd
from numpydoc_decorator import doc
from tqdm.auto import tqdm
from tqdm.dask import TqdmCallback

from ..util import (
    CacheMiss,
    LoggingHelper,
    check_colab_location,
    check_types,
    hash_params,
    init_filesystem,
)
from . import base_params


class AnophelesBase:
    def __init__(
        self,
        *,
        url: str,
        config_path: str,
        pre: bool,
        gcs_url: Optional[str],  # only used for colab location check
        major_version_number: int,
        major_version_path: str,
        bokeh_output_notebook: bool = False,
        log: Optional[Union[str, IO]] = None,
        debug: bool = False,
        show_progress: bool = False,
        check_location: bool = False,
        storage_options: Optional[Mapping] = None,
        results_cache: Optional[str] = None,
    ):
        self._url = url
        self._config_path = config_path
        self._pre = pre
        self._gcs_url = gcs_url
        self._major_version_number = major_version_number
        self._major_version_path = major_version_path
        self._debug = debug
        self._show_progress = show_progress

        # Set up logging.
        self._log = LoggingHelper(name=__name__, out=log, debug=debug)

        # Set up fsspec filesystem. N.B., we use fsspec here to allow for
        # accessing different types of storage - fsspec will automatically
        # detect which type of storage to use based on the URL provided.
        # E.g., if the URL begins with "gs://" then a GCSFileSystem will
        # be used to read from Google Cloud Storage.
        if storage_options is None:
            storage_options = dict()
        self._fs, self._base_path = init_filesystem(url, **storage_options)

        # Lazily load config.
        self._config: Optional[Dict] = None

        # Get bokeh to output plots to the notebook - this is a common gotcha,
        # users forget to do this and wonder why bokeh plots don't show.
        if bokeh_output_notebook:
            bokeh.io.output_notebook(hide_banner=True)

        # Check colab location is in the US.
        if check_location and self._gcs_url is not None:
            self._client_details = check_colab_location(
                gcs_url=self._gcs_url, url=self._url
            )
        else:
            self._client_details = None

        # Set up cache attributes.
        self._cache_releases: Optional[Tuple[str, ...]] = None
        self._cache_sample_sets: Dict[str, pd.DataFrame] = dict()
        self._cache_sample_set_to_release: Optional[Dict[str, str]] = None
        self._cache_files: Dict[str, bytes] = dict()

        # Set up results cache directory path.
        self._results_cache: Optional[Path] = None
        if results_cache is not None:
            self._results_cache = Path(results_cache).expanduser().resolve()

    def _progress(self, iterable, **kwargs):
        # progress doesn't mix well with debug logging
        disable = self._debug or not self._show_progress
        return tqdm(iterable, disable=disable, **kwargs)

    def _dask_progress(self, **kwargs):
        disable = not self._show_progress
        return TqdmCallback(disable=disable, **kwargs)

    @check_types
    def open_file(self, path: str) -> IO:
        full_path = f"{self._base_path}/{path}"
        return self._fs.open(full_path)

    @check_types
    def read_files(
        self,
        paths: Iterable[str],
        on_error: Literal["raise", "omit", "return"] = "return",
    ) -> Mapping[str, Union[bytes, Exception]]:
        # Check for any cached files.
        files = {
            path: data for path, data in self._cache_files.items() if path in paths
        }
        paths_not_cached = [p for p in paths if p not in self._cache_files]

        if paths_not_cached:
            # Prepend the base path.
            prefix = self._base_path + "/"
            full_paths = [prefix + path for path in paths_not_cached]

            # Retrieve all files. N.B., depending on what type of storage is
            # being used, the cat() function may be able to read multiple files
            # concurrently. E.g., this is true if the file system is a
            # GCSFileSystem, which achieves concurrency by using async I/O under the
            # hood. This is a useful performance optimisation, because reading a
            # file from GCS incurs some latency, and so reading many small files
            # from GCS can be slow if files are not read concurrently. Hence we
            # want to make use of cat() here and provide paths for all files to
            # read concurrently. For more information see:
            # https://filesystem-spec.readthedocs.io/en/latest/async.html
            full_path_files = self._fs.cat(full_paths, on_error=on_error)

            # Strip off the prefix.
            retrieved_files = {
                k.split(prefix, 1)[1]: v for k, v in full_path_files.items()
            }

            # Update the cache.
            self._cache_files.update(retrieved_files)

            # Add retrieved files to the result.
            files.update(retrieved_files)

        return files

    @property
    def config(self) -> Dict:
        if self._config is None:
            with self.open_file(self._config_path) as f:
                self._config = json.load(f)
        return self._config.copy()

    # Note regarding release identifiers and storage paths. Within the
    # data storage, we have used path segments like "v3", "v3.1", "v3.2",
    # etc., to separate data from different releases. There is an inconsistency
    # in this convention, because the "v3" should have been "v3.0". To
    # make the API more consistent, we would like to use consistent release
    # identifiers like "3.0", "3.1", "3.2", etc., as parameter values and
    # when release identifiers are added to returned dataframes. In order to
    # achieve this, below we define two functions that allow mapping between
    # these consistent release identifiers, and the less consistent release
    # storage path segments.

    @doc(
        summary="""
            Compatibility function, allows us to use release identifiers like "3.0"
            and "3.1" in the public API, and map these internally into storage path
            segments.
        """,
    )
    def _release_to_path(self, release: base_params.release) -> str:
        if release == f"{self._major_version_number}.0":
            # special case
            return self._major_version_path
        elif isinstance(release, str) and release.startswith(
            f"{self._major_version_number}."
        ):
            return f"v{release}"
        else:
            raise ValueError(f"Invalid release: {release!r}")

    @doc(
        summary="""
            Compatibility function, allows us to use release identifiers like "3.0"
            and "3.1" in the public API, and map these internally into storage path
            segments.
        """,
        parameters=dict(
            path="Path segment.",
        ),
    )
    def _path_to_release(self, path: str) -> str:
        if path == self._major_version_path:
            return f"{self._major_version_number}.0"
        elif path.startswith(f"v{self._major_version_number}."):
            return path[1:]
        else:
            raise RuntimeError(f"Unexpected release path: {path!r}")

    def _public_releases(self) -> Tuple[str, ...]:
        releases = tuple(self.config.get("PUBLIC_RELEASES", ()))
        return releases

    @doc(
        summary="""
            Here we discover which releases are available, by listing the storage
            directory and examining the subdirectories. This may include "pre-releases"
            where data may be incomplete.
        """
    )
    def _discover_releases(self) -> Tuple[str, ...]:
        sub_dirs = sorted([p.split("/")[-1] for p in self._fs.ls(self._base_path)])
        discovered_releases = tuple(
            sorted(
                [
                    self._path_to_release(d)
                    for d in sub_dirs
                    # FIXME: this matches v3 and v3.1, but also v3001.1
                    if d.startswith(f"v{self._major_version_number}")
                    and self._fs.exists(f"{self._base_path}/{d}/manifest.tsv")
                ]
            )
        )
        return discovered_releases

    @property
    def releases(self) -> Tuple[str, ...]:
        if self._cache_releases is None:
            if self._pre:
                self._cache_releases = self._discover_releases()
            else:
                self._cache_releases = self._public_releases()
        return self._cache_releases

    @property
    def client_location(self) -> str:
        details = self._client_details
        if details is not None:
            region = details.region
            country = details.country
            location = f"{region}, {country}"
        else:
            location = "unknown"
        return location

    def _read_sample_sets(self, *, single_release: str):
        """Read the manifest of sample sets for a single release."""
        # Construct a path for the manifest file.
        release_path = self._release_to_path(single_release)
        manifest_path = f"{release_path}/manifest.tsv"

        # Read the manifest into a pandas dataframe.
        with self.open_file(manifest_path) as f:
            df = pd.read_csv(f, sep="\t", na_values="")

        # Add a "release" column for convenience.
        df["release"] = single_release
        return df

    @check_types
    @doc(
        summary="Access a dataframe of sample sets",
        returns="A dataframe of sample sets, one row per sample set.",
    )
    def sample_sets(
        self,
        release: Optional[base_params.release] = None,
    ) -> pd.DataFrame:
        if release is None:
            # Retrieve sample sets from all available releases.
            release = self.releases

        if isinstance(release, str):
            # Retrieve sample sets for a single release.

            if release not in self.releases:
                raise ValueError(f"Release not available: {release!r}")

            try:
                df = self._cache_sample_sets[release]

            except KeyError:
                # Read and cache dataframe for performance.
                df = self._read_sample_sets(single_release=release)
                self._cache_sample_sets[release] = df

        elif isinstance(release, Sequence):
            # Ensure no duplicates.
            releases = sorted(set(release))

            # Retrieve and concatenate sample sets from multiple releases.
            df = pd.concat(
                [self.sample_sets(release=r) for r in releases],
                axis=0,
                ignore_index=True,
            )

        else:
            raise TypeError

        # Return copy to ensure cached dataframes aren't modified by user.
        return df.copy()

    @check_types
    @doc(
        summary="Find which release a sample set was included in.",
    )
    def lookup_release(self, sample_set: base_params.sample_set):
        if self._cache_sample_set_to_release is None:
            df_sample_sets = self.sample_sets().set_index("sample_set")
            self._cache_sample_set_to_release = df_sample_sets["release"].to_dict()

        try:
            return self._cache_sample_set_to_release[sample_set]
        except KeyError:
            raise ValueError(f"No release found for sample set {sample_set!r}")

    def _prep_sample_sets_param(
        self, *, sample_sets: Optional[base_params.sample_sets]
    ) -> List[str]:
        """Common handling for the `sample_sets` parameter. For convenience, we
        allow this to be a single sample set, or a list of sample sets, or a
        release identifier, or a list of release identifiers."""

        all_sample_sets = self.sample_sets()["sample_set"].to_list()

        if sample_sets is None:
            # All available sample sets.
            prepped_sample_sets = all_sample_sets

        elif isinstance(sample_sets, str):
            if sample_sets.startswith(f"{self._major_version_number}."):
                # Convenience, can use a release identifier to denote all sample sets in a release.
                prepped_sample_sets = self.sample_sets(release=sample_sets)[
                    "sample_set"
                ].to_list()

            else:
                # Single sample set, normalise to always return a list.
                prepped_sample_sets = [sample_sets]

        else:
            # Sequence of sample sets or releases.
            assert isinstance(sample_sets, Sequence)
            prepped_sample_sets = []
            for s in sample_sets:
                # Make a recursive call to handle the case where s is a release identifier.
                sp = self._prep_sample_sets_param(sample_sets=s)

                # Make sure we end up with a flat list of sample sets.
                prepped_sample_sets.extend(sp)

        # Ensure all sample sets selected at most once.
        prepped_sample_sets = sorted(set(prepped_sample_sets))

        # Check for bad sample sets.
        for s in prepped_sample_sets:
            if s not in all_sample_sets:
                raise ValueError(f"Sample set {s!r} not found.")

        return prepped_sample_sets

    def _results_cache_add_analysis_params(self, params: dict):
        # Expect sub-classes will override to add any analysis parameters.
        pass

    @check_types
    def results_cache_get(
        self, *, name: str, params: Dict[str, Any]
    ) -> Mapping[str, np.ndarray]:
        name = type(self).__name__.lower() + "_" + name
        if self._results_cache is None:
            raise CacheMiss
        params = params.copy()
        self._results_cache_add_analysis_params(params)
        cache_key, _ = hash_params(params)
        cache_path = self._results_cache / name / cache_key
        results_path = cache_path / "results.npz"
        if not results_path.exists():
            raise CacheMiss
        results = np.load(results_path)
        # TODO Do we need to read the arrays and then close the npz file?
        return results

    @check_types
    def results_cache_set(
        self, *, name: str, params: Dict[str, Any], results: Mapping[str, np.ndarray]
    ):
        name = type(self).__name__.lower() + "_" + name
        if self._results_cache is None:
            return
        params = params.copy()
        self._results_cache_add_analysis_params(params)
        cache_key, params_json = hash_params(params)
        cache_path = self._results_cache / name / cache_key
        cache_path.mkdir(exist_ok=True, parents=True)
        params_path = cache_path / "params.json"
        results_path = cache_path / "results.npz"
        with params_path.open(mode="w") as f:
            f.write(params_json)
        np.savez_compressed(results_path, **results)



"""Parameters for functions using plotly dash (e.g., haplotype networks)."""

from typing import Literal, Union

from typing_extensions import Annotated, TypeAlias

height: TypeAlias = Annotated[int, "Height of the Dash app in pixels (px)."]

width: TypeAlias = Annotated[Union[int, str], "Width of the Dash app."]

server_mode: TypeAlias = Annotated[
    Literal["inline", "external", "jupyterlab"],
    """
    Controls how the Jupyter Dash app will be launched. See
    https://medium.com/plotly/introducing-jupyterdash-811f1f57c02e for
    more information.
    """,
]

server_mode_default: server_mode = "inline"

server_port: TypeAlias = Annotated[
    int,
    "Manually override the port on which the Dash app will run.",
]



"""Parameters common to functions accessing haplotype data."""

from typing_extensions import Annotated, TypeAlias

analysis: TypeAlias = Annotated[
    str,
    """
    Which haplotype phasing analysis to use. See the
    `phasing_analysis_ids` property for available values.
    """,
]



"""Parameter definitions for Fst functions."""

from typing import Optional

from typing_extensions import Annotated, TypeAlias

from . import base_params

# N.B., window size can mean different things for different functions
window_size: TypeAlias = Annotated[
    int,
    "The size of windows (number of sites) used to calculate statistics within.",
]
cohort_size_default: Optional[base_params.cohort_size] = None
min_cohort_size_default: base_params.min_cohort_size = 15
max_cohort_size_default: base_params.max_cohort_size = 50



"""Parameters for haplotype network functions."""

from typing import List, Mapping

from typing_extensions import Annotated, TypeAlias

max_dist: TypeAlias = Annotated[
    int,
    "Join network components up to a maximum distance of 2 SNP differences.",
]

max_dist_default: max_dist = 2

color: TypeAlias = Annotated[
    str,
    """
    Identifies a column in the sample metadata which determines the colour
    of pie chart segments within nodes.
    """,
]

color_discrete_sequence: TypeAlias = Annotated[
    List, "Provide a list of colours to use."
]

color_discrete_map: TypeAlias = Annotated[
    Mapping, "Provide an explicit mapping from values to colours."
]

category_order: TypeAlias = Annotated[
    List,
    "Control the order in which values appear in the legend.",
]

node_size_factor: TypeAlias = Annotated[
    int,
    "Control the sizing of nodes.",
]

node_size_factor_default: node_size_factor = 50

layout: TypeAlias = Annotated[
    str,
    "Name of the network layout to use to position nodes.",
]

layout_default: layout = "cose"

layout_params: TypeAlias = Annotated[
    Mapping,
    "Additional parameters to the layout algorithm.",
]



import json
import os

import dask.array as da
import pandas as pd
import xarray
import zarr

from malariagen_data.util import (
    DIM_ALLELE,
    DIM_PLOIDY,
    DIM_SAMPLE,
    DIM_VARIANT,
    _prep_geneset_attributes_arg,
    da_from_zarr,
    init_filesystem,
    init_zarr_store,
    read_gff3,
    resolve_region,
    unpack_gff3_attributes,
)


class PlasmodiumDataResource:
    def __init__(
        self,
        data_config,
        url=None,
        **kwargs,
    ):
        # setup filesystem
        self.CONF = self._load_config(data_config)
        if not url:
            url = self.CONF["default_url"]
        self._fs, self._path = init_filesystem(url, **kwargs)

        # setup caches
        self._cache_sample_metadata = None
        self._cache_variant_calls_zarr = None
        self._cache_genome = None
        self._cache_genome_features = dict()

        self.extended_calldata_variables = self.CONF["extended_calldata_variables"]
        self.extended_variant_fields = self.CONF["extended_variant_fields"]
        self.contigs = self.CONF["reference_contigs"]

    def _load_config(self, data_config):
        """Load the config for data structure on the cloud into json format."""
        with open(data_config) as json_conf:
            config_content = json.load(json_conf)
        return config_content

    def sample_metadata(self):
        """Access sample metadata and return as pandas dataframe.
        Returns
        -------
        df : pandas.DataFrame
            A dataframe of sample metadata on the samples that were sequenced as part of this resource.
            Includes the time and place of collection, quality metrics, and accesion numbers.
            One row per sample.
        """
        if self._cache_sample_metadata is None:
            path = os.path.join(self._path, self.CONF["metadata_path"])
            with self._fs.open(path) as f:
                self._cache_sample_metadata = pd.read_csv(f, sep="\t", na_values="")
        return self._cache_sample_metadata

    def _open_variant_calls_zarr(self):
        """Open variant calls zarr.

        Returns
        -------
        root : zarr.hierarchy.Group
            Root of zarr containing information on variant calls.

        """
        if self._cache_variant_calls_zarr is None:
            path = os.path.join(self._path, self.CONF["variant_calls_zarr_path"])
            store = init_zarr_store(fs=self._fs, path=path)
            self._cache_variant_calls_zarr = zarr.open_consolidated(store=store)
        return self._cache_variant_calls_zarr

    def _add_coordinates(self, root, inline_array, chunks, var_names_for_outputs):
        """Add coordinate variables in zarr to dictionary"""
        # coordinates
        coords = dict()
        for var_name in ["POS", "CHROM"]:
            z = root[f"variants/{var_name}"]
            var = da_from_zarr(z, inline_array=inline_array, chunks=chunks)
            coords[f"variant_{var_names_for_outputs[var_name]}"] = [DIM_VARIANT], var

        z = root["samples"]
        sample_id = da_from_zarr(z, inline_array=inline_array, chunks=chunks)
        coords["sample_id"] = [DIM_SAMPLE], sample_id
        return coords

    def _add_default_data_vars(self, root, inline_array, chunks, var_names_for_outputs):
        """Add default set of variables from zarr to dictionary"""
        data_vars = dict()

        # Add variant_allele as combination of REF and ALT
        ref_z = root["variants/REF"]
        alt_z = root["variants/ALT"]
        ref = da_from_zarr(ref_z, inline_array=inline_array, chunks=chunks)
        alt = da_from_zarr(alt_z, inline_array=inline_array, chunks=chunks)
        variant_allele = da.concatenate([ref[:, None], alt], axis=1)
        data_vars["variant_allele"] = [DIM_VARIANT, DIM_ALLELE], variant_allele

        # other default variant values
        configurable_default_variant_variables = self.CONF["default_variant_variables"]
        for var_name in configurable_default_variant_variables:
            z = root[f"variants/{var_name}"]
            dimension = configurable_default_variant_variables[var_name]
            var = da_from_zarr(z, inline_array=inline_array, chunks=chunks)
            if var_name in var_names_for_outputs.keys():
                var_name = var_names_for_outputs[var_name]
            data_vars[f"variant_{var_name}"] = dimension, var

        # call arrays
        gt_z = root["calldata/GT"]
        ad_z = root["calldata/AD"]
        call_genotype = da_from_zarr(gt_z, inline_array=inline_array, chunks=chunks)
        call_ad = da_from_zarr(ad_z, inline_array=inline_array, chunks=chunks)
        data_vars["call_genotype"] = (
            [DIM_VARIANT, DIM_SAMPLE, DIM_PLOIDY],
            call_genotype,
        )
        data_vars["call_AD"] = ([DIM_VARIANT, DIM_SAMPLE, DIM_ALLELE], call_ad)

        return data_vars

    def _add_extended_data(self, root, inline_array, chunks, data_vars):
        """Add all variables not included in default set to dictionary"""

        subset_extended_variants = self.extended_variant_fields
        subset_extended_calldata = self.extended_calldata_variables

        for var_name in subset_extended_calldata:
            z = root[f"calldata/{var_name}"]
            var = da_from_zarr(z, inline_array=inline_array, chunks=chunks)
            data_vars[f"call_{var_name}"] = (
                subset_extended_calldata[var_name],
                var,
            )

        for var_name in subset_extended_variants:
            z = root[f"variants/{var_name}"]
            field = da_from_zarr(z, inline_array=inline_array, chunks=chunks)
            data_vars[f"variant_{var_name}"] = (
                subset_extended_variants[var_name],
                field,
            )
        return data_vars

    def variant_calls(self, extended=False, inline_array=True, chunks="native"):
        """Access variant sites, site filters and genotype calls.

        Parameters
        ----------
        extended : bool, optional
            If False only the default variables are returned. If True all variables from the zarr are returned. Defaults to False.
        inline_array : bool, optional
            Passed through to dask.array.from_array(). Defaults to True.
        chunks : str, optional
            If 'auto' let dask decide chunk size. If 'native' use native zarr chunks.
            Also can be a target size, e.g., '200 MiB'. Defaults to "native".

        Returns
        -------
        ds : xarray.Dataset
            Dataset containing either default or extended variables from the variant calls Zarr.
        """
        # setup
        root = self._open_variant_calls_zarr()
        var_names_for_outputs = {
            "POS": "position",
            "CHROM": "chrom",
            "FILTER_PASS": "filter_pass",
        }

        # Add default data
        coords = self._add_coordinates(
            root, inline_array, chunks, var_names_for_outputs
        )
        data_vars = self._add_default_data_vars(
            root, inline_array, chunks, var_names_for_outputs
        )

        # Add extended data
        if extended:
            data_vars = self._add_extended_data(root, inline_array, chunks, data_vars)

        # create a dataset
        ds = xarray.Dataset(data_vars=data_vars, coords=coords)

        return ds

    def open_genome(self):
        """Open the reference genome zarr.

        Returns
        -------
        root : zarr.hierarchy.Group
            Zarr hierarchy containing the reference genome sequence.

        """
        if self._cache_genome is None:
            path = os.path.join(self._path, self.CONF["reference_path"])
            store = init_zarr_store(fs=self._fs, path=path)
            self._cache_genome = zarr.open_consolidated(store=store)
        return self._cache_genome

    def resolve_region(self, region):
        """Convert a genome region into a standard data structure.

        Parameters
        ----------
        region: str
            Chromosome arm (e.g., "2L"), gene name (e.g., "AGAP007280") or
            genomic region defined with coordinates (e.g.,
            "2L:44989425-44998059").

        Returns
        -------
        out : Region
            A named tuple with attributes contig, start and end.

        """

        return resolve_region(self, region)

    def _subset_genome_sequence_region(
        self, genome, region, inline_array=True, chunks="native"
    ):
        """Sebset reference genome sequence."""
        region = self.resolve_region(region)
        z = genome[region.contig]

        d = da_from_zarr(z, inline_array=inline_array, chunks=chunks)

        if region.start:
            slice_start = region.start - 1
        else:
            slice_start = None
        if region.end:
            slice_stop = region.end
        else:
            slice_stop = None
        loc_region = slice(slice_start, slice_stop)

        return d[loc_region]

    def genome_sequence(self, region="*", inline_array=True, chunks="native"):
        """Access the reference genome sequence.

        Parameters
        ----------
        region: str or list of str or Region or list of Region. Defaults to '*'
            Chromosome (e.g., "Pf3D7_07_v3"), gene name (e.g., "PF3D7_0709000"), genomic
            region defined with coordinates (e.g., "Pf3D7_07_v3:1-500").
            Multiple values can be provided as a list, in which case data will
            be concatenated, e.g., ["Pf3D7_07_v3:1-500","Pf3D7_02_v3:15-20","Pf3D7_03_v3:40-50"].
        inline_array : bool, optional
            Passed through to dask.array.from_array().
        chunks : str, optional
            If 'auto' let dask decide chunk size. If 'native' use native zarr
            chunks. Also, can be a target size, e.g., '200 MiB'.

        Returns
        -------
        d : dask.array.Array
            An array of nucleotides giving the reference genome sequence for the
            given region/gene/contig.

        """
        genome = self.open_genome()
        if type(region) not in [tuple, list] and region != "*" and region is not None:
            d = self._subset_genome_sequence_region(
                genome=genome,
                region=region,
                inline_array=inline_array,
                chunks=chunks,
            )
        else:
            region = tuple(region)
            if region == tuple("*"):
                region = self.contigs
            d = da.concatenate(
                [
                    self._subset_genome_sequence_region(
                        genome=genome,
                        region=r,
                        inline_array=inline_array,
                        chunks=chunks,
                    )
                    for r in region
                ]
            )
        return d

    def genome_features(self, attributes=("ID", "Parent", "Name", "alias")):
        """Access genome feature annotations.

        Parameters
        ----------
        attributes : list of str, optional
            Attribute keys to unpack into columns. Provide "*" to unpack all attributes.

        Returns
        -------
        df : pandas.DataFrame

        """
        # Attributes
        attributes = _prep_geneset_attributes_arg(attributes)

        try:
            df = self._cache_genome_features[attributes]
        except KeyError:
            path = os.path.join(self._path, self.CONF["annotations_path"])
            with self._fs.open(path, mode="rb") as f:
                df = read_gff3(f, compression="gzip")
            if attributes is not None:
                df = unpack_gff3_attributes(df, attributes=attributes)
            self._cache_genome_features[attributes] = df

        return df



"""Parameter definitions for XPEHH analysis functions."""

from typing import Tuple, Union

import numpy as np
from typing_extensions import Annotated, TypeAlias

from . import base_params

window_size: TypeAlias = Annotated[
    int,
    """
        The size of window in number of SNPs used to summarise XP-EHH over.
        If None, per-variant XP-EHH values are returned.
        """,
]
window_size_default: window_size = 200
min_cohort_size_default: base_params.min_cohort_size = 15
max_cohort_size_default: base_params.max_cohort_size = 50
percentiles: TypeAlias = Annotated[
    Union[int, Tuple[int, ...]],
    """
    If window size is specified, this returns the XP-EHH percentiles
    for each window.
    """,
]
percentiles_default: percentiles = (50, 75, 100)
filter_min_maf: TypeAlias = Annotated[
    float,
    """
    Minimum minor allele frequency to use for filtering prior to passing
    haplotypes to allel.xpehh function
    """,
]
filter_min_maf_default: filter_min_maf = 0.05
map_pos: TypeAlias = Annotated[
    np.ndarray,
    """
    Variant positions (genetic map distance).
    """,
]
min_ehh: TypeAlias = Annotated[
    float,
    """
    Minimum EHH beyond which to truncate integrated haplotype homozygosity
    calculation.
    """,
]
min_ehh_default: min_ehh = 0.05
max_gap: TypeAlias = Annotated[
    int,
    """
    Do not report scores if EHH spans a gap larger than this number of
    base pairs.
    """,
]
max_gap_default: max_gap = 200_000
gap_scale: TypeAlias = Annotated[
    int, "Rescale distance between variants if gap is larger than this value."
]
gap_scale_default: gap_scale = 20_000
include_edges: TypeAlias = Annotated[
    bool,
    """
    If True, report scores even if EHH does not decay below min_ehh at the
    end of the chromosome.
    """,
]
use_threads: TypeAlias = Annotated[
    bool, "If True, use multiple threads to compute XP-EHH."
]
palette: TypeAlias = Annotated[
    str, "Name of bokeh palette to use for plotting multiple percentiles."
]
palette_default: palette = "Blues"



"""Parameters for PCA functions."""

import numpy as np
import pandas as pd
from typing_extensions import Annotated, TypeAlias

n_snps: TypeAlias = Annotated[
    int,
    """
    The desired number of SNPs to use when running the analysis.
    SNPs will be evenly thinned to approximately this number.
    """,
]

thin_offset: TypeAlias = Annotated[
    int,
    """
    Starting index for SNP thinning. Change this to repeat the analysis
    using a different set of SNPs.
    """,
]

thin_offset_default: thin_offset = 0

min_minor_ac: TypeAlias = Annotated[
    int,
    """
    The minimum minor allele count. SNPs with a minor allele count
    below this value will be excluded prior to thinning.
    """,
]

min_minor_ac_default: min_minor_ac = 2

max_missing_an: TypeAlias = Annotated[
    int,
    """
    The maximum number of missing allele calls to accept. SNPs with
    more than this value will be excluded prior to thinning. Set to 0
    (default) to require no missing calls.
    """,
]

max_missing_an_default = 0

n_components: TypeAlias = Annotated[
    int,
    "Number of components to return.",
]

n_components_default: n_components = 20

df_pca: TypeAlias = Annotated[
    pd.DataFrame,
    """
    A dataframe of sample metadata, with columns "PC1", "PC2", "PC3",
    etc., added.
    """,
]

evr: TypeAlias = Annotated[
    np.ndarray,
    "An array of explained variance ratios, one per component.",
]



from typing import Tuple

import dask.array as da
import zarr
from numpydoc_decorator import doc

from ..util import (
    Region,
    check_types,
    da_from_zarr,
    init_zarr_store,
    parse_single_region,
)
from . import base_params
from .base import AnophelesBase


class AnophelesGenomeSequenceData(AnophelesBase):
    def __init__(self, **kwargs):
        # N.B., this class is designed to work cooperatively, and
        # so it's important that any remaining parameters are passed
        # to the superclass constructor.
        super().__init__(**kwargs)

        # Initialize cache attributes.
        self._cache_genome = None

    @property
    def contigs(self) -> Tuple[str, ...]:
        return tuple(self.config["CONTIGS"])

    @property
    def _genome_zarr_path(self) -> str:
        return self.config["GENOME_ZARR_PATH"]

    @property
    def _genome_fasta_path(self) -> str:
        return self.config["GENOME_FASTA_PATH"]

    @property
    def _genome_fai_path(self) -> str:
        return self.config["GENOME_FAI_PATH"]

    @property
    def _genome_ref_id(self) -> str:
        return self.config["GENOME_REF_ID"]

    @property
    def _genome_ref_name(self) -> str:
        return self.config["GENOME_REF_NAME"]

    @check_types
    @doc(
        summary="Open the reference genome zarr.",
        returns="Zarr hierarchy containing the reference genome sequence.",
    )
    def open_genome(self) -> zarr.hierarchy.Group:
        if self._cache_genome is None:
            path = f"{self._base_path}/{self._genome_zarr_path}"
            store = init_zarr_store(fs=self._fs, path=path)
            self._cache_genome = zarr.open_consolidated(store=store)
        return self._cache_genome

    def _genome_sequence_for_contig(self, *, contig, inline_array, chunks):
        """Obtain the genome sequence for a given contig as an array."""
        assert contig in self.contigs
        root = self.open_genome()
        z = root[contig]
        d = da_from_zarr(z, inline_array=inline_array, chunks=chunks)
        return d

    @check_types
    @doc(
        summary="Access the reference genome sequence.",
        returns="""
            An array of nucleotides giving the reference genome sequence for the
            given contig.
        """,
    )
    def genome_sequence(
        self,
        region: base_params.single_region,
        inline_array: base_params.inline_array = base_params.inline_array_default,
        chunks: base_params.chunks = base_params.chunks_default,
    ) -> da.Array:
        # Parse the region parameter into a Region object.
        resolved_region: Region = parse_single_region(self, region)
        del region

        # Obtain complete sequence for the requested contig.
        d = self._genome_sequence_for_contig(
            contig=resolved_region.contig, inline_array=inline_array, chunks=chunks
        )

        # Deal with region start and stop.
        slice_start = slice_stop = None
        if resolved_region.start:
            slice_start = resolved_region.start - 1
        if resolved_region.end:
            slice_stop = resolved_region.end
        loc_region = slice(slice_start, slice_stop)

        return d[loc_region]



"""Parameters for haplotype clustering functions."""

from typing import Literal

from typing_extensions import Annotated, TypeAlias

linkage_method: TypeAlias = Annotated[
    Literal["single", "complete", "average", "weighted", "centroid", "median", "ward"],
    """
    The linkage algorithm to use. See the Linkage Methods section of the
    scipy.cluster.hierarchy.linkage docs for full descriptions.
    """,
]

linkage_method_default: linkage_method = "single"

count_sort: TypeAlias = Annotated[
    bool,
    """
    For each node n, the order (visually, from left-to-right) n's two descendant
    links are plotted is determined by this parameter. If True, the child with
    the minimum number of original objects in its cluster is plotted first. Note
    distance_sort and count_sort cannot both be True.
    """,
]

distance_sort: TypeAlias = Annotated[
    bool,
    """
    For each node n, the order (visually, from left-to-right) n's two descendant
    links are plotted is determined by this parameter. If True, The child with the
    minimum distance between its direct descendants is plotted first.
    """,
]
